#!/usr/bin/perl
#-------------------------------------------------------------------------------
# unicc_bor.pl
#
# Copyright (C) 2006-2008 Oliver Hamann.
#
# Homepage: http://eaglemode.sourceforge.net/
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3 as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License version 3 for
# more details.
#
# You should have received a copy of the GNU General Public License version 3
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#-------------------------------------------------------------------------------

use strict;
use warnings;
use File::Basename;
use File::Spec::Functions;
use File::stat;


#=============================== Parse arguments ===============================

my $Type="cexe";
my $Name="unnamed";
my $ObjDir=".";
my $BinDir=".";
my $LibDir=".";
my @LibSearchDirs=();
my @IncSearchDirs=();
my @Links=();
my $Math=0;
my $Rtti=0;
my $Exceptions=0;
my $Debug=0;
my @Defs=();
my @Srcs=();

for (my $i=0; $i<@ARGV; $i++) {
	if ($ARGV[$i] eq "--type" and $i+1<@ARGV) {
		$Type=$ARGV[++$i];
	}
	elsif ($ARGV[$i] eq "--name" and $i+1<@ARGV) {
		$Name=$ARGV[++$i];
	}
	elsif ($ARGV[$i] eq "--obj-dir" and $i+1<@ARGV) {
		$ObjDir=$ARGV[++$i];
	}
	elsif ($ARGV[$i] eq "--bin-dir" and $i+1<@ARGV) {
		$BinDir=$ARGV[++$i];
	}
	elsif ($ARGV[$i] eq "--lib-dir" and $i+1<@ARGV) {
		$LibDir=$ARGV[++$i];
		push(@LibSearchDirs,$LibDir);
	}
	elsif ($ARGV[$i] eq "--lib-search-dir" and $i+1<@ARGV) {
		push(@LibSearchDirs,$ARGV[++$i]);
	}
	elsif ($ARGV[$i] eq "--inc-search-dir" and $i+1<@ARGV) {
		push(@IncSearchDirs,$ARGV[++$i]);
	}
	elsif ($ARGV[$i] eq "--link" and $i+1<@ARGV) {
		push(@Links,$ARGV[++$i]);
	}
	elsif ($ARGV[$i] eq "--math") {
		$Math=1;
	}
	elsif ($ARGV[$i] eq "--rtti") {
		$Rtti=1;
	}
	elsif ($ARGV[$i] eq "--exceptions") {
		$Exceptions=1;
	}
	elsif ($ARGV[$i] eq "--debug") {
		$Debug=1;
	}
	elsif ($ARGV[$i] eq "--def" and $i+1<@ARGV) {
		push(@Defs,$ARGV[++$i]);
	}
	elsif ($ARGV[$i]=~/^--/) {
		die("Unknown option '$ARGV[$i]', stopped");
	}
	else {
		push(@Srcs,$ARGV[$i]);
	}
}

if (@Srcs <= 0) {
	die("No source files, stopped");
}


#====================== Translate '/' to '\' in all paths ======================

$ObjDir=~tr/\//\\/;
$BinDir=~tr/\//\\/;
$LibDir=~tr/\//\\/;
for (my $i=0; $i<@LibSearchDirs; $i++) { $LibSearchDirs[$i]=~tr/\//\\/; }
for (my $i=0; $i<@IncSearchDirs; $i++) { $IncSearchDirs[$i]=~tr/\//\\/; }
for (my $i=0; $i<@Srcs; $i++) { $Srcs[$i]=~tr/\//\\/; }


#================= Have paths of output files and other infos ==================

my @CSuffixes;
my @CppSuffixes;
my @Objs=();
my @IsCpp=();
my $AnyCpp;
my $Output;
my $TDSFile;
my $DefFile;
my $ImportLib;

@CSuffixes=(".c", ".C");
@CppSuffixes=(".cpp", ".cxx", ".cc");
$AnyCpp=0;
for (my $i=0; $i<@Srcs; $i++) {
	my ($f,$d,$s)=fileparse($Srcs[$i],@CSuffixes);
	if ($s eq "") {
		($f,$d,$s)=fileparse($Srcs[$i],@CppSuffixes);
		if ($s eq "") {
			die("Source file '$Srcs[$i]' has illegal suffix, stopped");
		}
		push(@IsCpp,1);
		$AnyCpp=1;
	}
	else {
		push(@IsCpp,0);
	}
	push(@Objs,catfile($ObjDir,"$f.obj"));
}

if ($Type eq "cexe" or $Type eq "wexe") {
	$Output=catfile($BinDir,"$Name.exe");
	$TDSFile=catfile($BinDir,"$Name.tds");
}
elsif ($Type eq "lib") {
	$Output=catfile($LibDir,"$Name.lib");
	$TDSFile=catfile($BinDir,"$Name.tds");
}
elsif ($Type eq "dynlib") {
	$Output=catfile($LibDir,"$Name.dll");
	$TDSFile=catfile($LibDir,"$Name.tds");
	$DefFile=catfile($ObjDir,"$Name.def");
	$ImportLib=catfile($LibDir,"$Name.lib");
}
else {
	die("Illegal type of output: '$Type', stopped");
}


#==================== GetDirectIncludes & GetSrcUpdateTime =====================

my %DirectIncludesCache;
my %SrcUpdateTimeCache;

sub GetDirectIncludes
{
	my $srcFile=shift;
	my @result;
	my $fh;

	if (exists $DirectIncludesCache{$srcFile}) {
		return @{$DirectIncludesCache{$srcFile}};
	}
	@result=();
	open($fh,$srcFile);
	while (defined(my $s=readline($fh))) {
		if (index($s,"#")<0) { next; }
		while (ord($s)<=32) { $s=substr($s,1); }
		if (substr($s,0,1) ne "#") { next; }
		$s=substr($s,1);
		while (ord($s)<=32) { $s=substr($s,1); }
		if (substr($s,0,7) ne "include") { next; }
		$s=substr($s,7);
		while (ord($s)<=32) { $s=substr($s,1); }
		my $e;
		if (substr($s,0,1) eq "<") { $e=">"; }
		elsif (substr($s,0,1) eq "\"") { $e="\""; }
		else { next; }
		my $i=index($s,$e,2);
		if ($i<2) { next; }
		$s=substr($s,1,$i-1);
		my @dirs=@IncSearchDirs;
		if ($e eq "\"") {
			my ($f,$d)=fileparse($s);
			@dirs=($d,@dirs);
		}
		for (my $i=0; $i<@dirs; $i++) {
			my $p=catfile($dirs[$i],$s);
			if (! -f $p) { next; }
			push(@result,$p);
			last;
		}
	}
	close($fh);
	$DirectIncludesCache{$srcFile}=[@result];
	return @result;
}

sub GetSrcUpdateTime
{
	my $srcFile=shift;
	my $r;

	$r=$SrcUpdateTimeCache{$srcFile};
	if (defined($r)) {
		if ($r!=-1) { return $r; }
		my @a=($srcFile);
		for (my $i=0; $i<@a; $i++) {
			my $s=stat($a[$i]);
			if ($i==0 or $r<$s->mtime) { $r=$s->mtime; }
			my @d=GetDirectIncludes($a[$i]);
			for (my $j=0; $j<@d; $j++) {
				my $k;
				for ($k=0; $k<@a; $k++) {
					if ($d[$j] eq $a[$k]) { last; }
				}
				if ($k>=@a) { push(@a,$d[$j]); }
			}
		}
	}
	else {
		$SrcUpdateTimeCache{$srcFile}=-1;
		my $s=stat($srcFile);
		$r=$s->mtime;
		my @d=GetDirectIncludes($srcFile);
		for (my $i=0; $i<@d; $i++) {
			my $t=GetSrcUpdateTime($d[$i]);
			if ($r<$t) { $r=$t }
		}
	}
	$SrcUpdateTimeCache{$srcFile}=$r;
	return $r;
}


#==================== Compile source files to object files =====================

for (my $i=0; $i<@Srcs; $i++) {
	if (-f $Objs[$i]) {
		my $s=stat($Objs[$i]);
		if ($s->mtime>=GetSrcUpdateTime($Srcs[$i])) { next; }
	}
	my @args=();
	push(@args,"bcc32");
	push(@args,"-q");
	push(@args,"-O2");
	if ($Debug) {
		push(@args,"-y");
		push(@args,"-v");
	}
	if ($Rtti) { push(@args,"-RT"); }
	else { push(@args,"-RT-"); }
	if ($Exceptions) { push(@args,"-x"); }
	else { push(@args,"-x-"); }
	push(@args,"-j4");
	push(@args,"-w");
	push(@args,"-w-par");
	push(@args,"-w-aus");
	if ($Type eq "cexe") { push(@args,"-tWC"); }
	if ($Type eq "wexe") { push(@args,"-tW"); }
	if ($Type eq "dynlib") { push(@args,"-tWD"); }
	push(@args,"-tWM"); # multi-threaded
	push(@args,"-tWR");
	for (my $j=0; $j<@IncSearchDirs; $j++) {
		push(@args,"-I$IncSearchDirs[$j]");
	}
	for (my $j=0; $j<@Defs; $j++) {
		push(@args,"-D$Defs[$j]");
	}
	my ($f,$d)=fileparse($Objs[$i]);
	push(@args,"-o$Objs[$i]");
	push(@args,"-c");
	push(@args,$Srcs[$i]); # Must be the last
	print("@args\n");
	if (system(@args)!=0) { exit(1); }
}


#=========================== Output file up-to-date? ===========================

if (-e $Output) {
	my $t=0;
	for (my $i=0; $i<@Objs; $i++) {
		my $s=stat($Objs[$i]);
		if ($t<$s->mtime) { $t=$s->mtime; }
	}
	for (my $i=0; $i<@Links; $i++) {
		for (my $j=0; $j<@LibSearchDirs; $j++) {
			my $p;
			$p=catfile($LibSearchDirs[$j],"$Links[$i].lib");
			if (-e $p) {
				my $s=stat($p);
				if ($t<$s->mtime) { $t=$s->mtime; }
			}
		}
	}
	my $s=stat($Output);
	if ($s->mtime>=$t) { exit(0); }
}


#============================== Produce def file ===============================

if ($Type eq "dynlib") {
	print("Creating DEF-File: $DefFile\n");
	my $dfh;
	open($dfh,">",$DefFile);
	print($dfh "LIBRARY $Name\n");
	print($dfh "EXPORTS\n");
	my %exports;
	my $n=0;
	for (my $i=0; $i<@Objs; $i++) {
		my @dump=`tdump -q -m $Objs[$i]`;
		if ($? != 0) { exit(1); }
		for (my $j=0; $j<@dump; $j++) {
			my $str=$dump[$j];
			if (index($str,"virtual(_TEXT)")<0) { next; }
			while (ord($str)<=32) { $str=substr($str,1); }
			if (index($str,"Name:")!=0) { next; }
			$str=substr($str,5);
			while (ord($str)<=32) { $str=substr($str,1); }
			my $k=ord($str);
			if ($k<48 or $k>57) { next; }
			for (;;) {
				$str=substr($str,1);
				$k=ord($str);
				if ($k<48 or $k>57) { last; }
			}
			if (index($str,":")!=0) { next; }
			$str=substr($str,1);
			while (ord($str)<=32) { $str=substr($str,1); }
			if (index($str,"'")!=0) { next; }
			$str=substr($str,1);
			$k=index($str,"'");
			if ($k<0) { next; }
			my $sym=substr($str,0,$k);
			$str=substr($str,$k+1);
			while (ord($str)<=32) { $str=substr($str,1); }
			if (index($str,"virtual(_TEXT)")!=0) { next; }
			my $sym2=$sym;
			if (substr($sym,0,1) eq "_") { $sym2=substr($sym,1); }
			if (exists $exports{$sym}) { next; }
			$exports{$sym}=1;
			$n++;
			print($dfh " $sym \@$n\n");
			if ($sym2 ne $sym) {
				$n++;
				print($dfh " $sym2=$sym \@$n\n");
			}
		}
	}
	close($dfh);
}


#============================= Produce output file =============================

# Delete old output file.
if (-e $Output) { unlink($Output); }

# Create new one.
if ($Type eq "cexe" or $Type eq "wexe" or $Type eq "dynlib") {
	my @args=();
	push(@args,"ilink32");
	push(@args,"-q");
	push(@args,"-w");
	push(@args,"-x");
	push(@args,"-Gn");
	if ($Type eq "cexe") {
		push(@args,"-ap");
		push(@args,"-Tpe");
	}
	if ($Type eq "wexe") {
		push(@args,"-aa");
		push(@args,"-Tpe");
	}
	if ($Type eq "dynlib") {
		push(@args,"-Tpd");
	}
	push(@args,"-S:2048576"); #??? stack size
	if ($Debug) { push(@args,"-v"); }
	for (my $i=0; $i<@LibSearchDirs; $i++) {
		push(@args,"-L$LibSearchDirs[$i]");
	}
	if ($Type eq "cexe")   { push(@args,"c0x32.obj"); }
	if ($Type eq "wexe")   { push(@args,"c0w32.obj"); }
	if ($Type eq "dynlib") { push(@args,"c0d32.obj"); }
	for (my $i=0; $i<@Objs; $i++) {
		push(@args,$Objs[$i]);
	}
	push(@args,",");
	push(@args,$Output);
	push(@args,",");
	push(@args,",");
	push(@args,"cw32mti.lib"); # or cw32i.lib if not multi-threaded
	push(@args,"import32.lib");
	for (my $i=0; $i<@Links; $i++) {
		if (
			$Links[$i] eq "user32" or
			$Links[$i] eq "advapi32" or
			$Links[$i] eq "shell32" or
			$Links[$i] eq "gdi32"
			#??? to be continued
		) { next; }
		push(@args,"$Links[$i].lib");
	}
	push(@args,",");
	if ($Type eq "dynlib") { push(@args,$DefFile); }
	push(@args,",");
	print("@args\n");
	if (system(@args)!=0) { exit(1); }
	if ($Type eq "dynlib") {
		@args=();
		push(@args,"implib");
		push(@args,"$ImportLib");
		push(@args,"$Output");
		print("@args\n");
		if (system(@args)!=0) { exit(1); }
	}
	if (!$Debug) {
		if (-e $TDSFile) {
			print("Deleting $TDSFile\n");
			unlink($TDSFile);
		}
	}
}
elsif ($Type eq "lib") {
	my @args=();
	push(@args,"tlib");
	push(@args,"/P64");
	push(@args,"$Output");
	for (my $i=0; $i<@Objs; $i++) {
		push(@args,"+$Objs[$i]");
	}
	print("@args\n");
	if (system(@args)!=0) { exit(1); }
}
else {
	die("Illegal type of output: '$Type', stopped");
}
